@@ -35,8 +35,10 @@ from storm.expr import (
     compile_sql_token)
 from storm.variables import Variable, ListVariable, RawStrVariable
 from storm.database import Database, Connection, Result
-from storm.exceptions import (install_exceptions, DatabaseModuleError,
-                              OperationalError, ProgrammingError)
+from storm.exceptions import (
+    install_exceptions, DatabaseError, DatabaseModuleError,
+    OperationalError, ProgrammingError, TimeoutError)
+from storm.tracer import TimeoutTracer
 
 
 install_exceptions(psycopg2)
@@ -75,7 +77,7 @@ def compile_currval(compile, expr, state):
     """Compile a currval.
 
     This is a bit involved because we have to get escaping right.  Here
-    are a few cases to keep in mind:
+    are a few cases to keep in mind::
 
         currval('thetable_thecolumn_seq')
         currval('theschema.thetable_thecolumn_seq')
@@ -196,7 +198,7 @@ def compile_str_variable_with_E(compile, variable, state):
 
     PostgreSQL 8.2 will issue a warning without it, and psycopg
     will use the plain '' rather than E''.  The problem is being
-    tracked at the following URL:
+    tracked at the following URL::
 
         http://www.initd.org/tracker/psycopg/ticket/202
 
@@ -237,7 +239,7 @@ class PostgresConnection(Connection):
         in-memory objects with their specific rows.
         """
         if (isinstance(statement, Insert) and
-            self._database._version >= (8, 2) and
+            self._database._version >= 80200 and
             statement.primary_variables is not Undef and
             statement.primary_columns is not Undef):
 
@@ -290,23 +292,40 @@ class PostgresConnection(Connection):
         # I have no idea why I am seeing the last exception message
         # after upgrading to Gutsy.
         msg = str(exc)
-        return (msg.startswith("server closed the connection unexpectedly") or
-                msg.startswith("could not connect to server") or
-                msg.startswith("no connection to the server") or
-                msg.startswith("connection not open") or
-                msg.startswith("losed the connection unexpectedly"))
+        return ("server closed the connection unexpectedly" in msg or
+                "could not connect to server" in msg or
+                "no connection to the server" in msg or
+                "connection not open" in msg or
+                "losed the connection unexpectedly" in msg)
 
 
 class Postgres(Database):
 
     connection_factory = PostgresConnection
 
+    # An integer representing the server version.  If the server does
+    # not support the server_version_num variable, this will be set to
+    # 0.  In practice, this means the variable will be 0 or greater
+    # than or equal to 80200.
     _version = None
 
     def __init__(self, uri):
         if psycopg2 is dummy:
             raise DatabaseModuleError("'psycopg2' module not found")
         self._dsn = make_dsn(uri)
+        isolation = uri.options.get("isolation", "serializable")
+        isolation_mapping = {
+            "autocommit": psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT,
+            "serializable": psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE,
+            "read-committed": psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED,
+        }
+        try:
+            self._isolation = isolation_mapping[isolation]
+        except KeyError:
+            raise ValueError(
+                "Unknown serialization level %r: expected one of "
+                "'autocommit', 'serializable', 'read-committed'" %
+                (isolation,))
 
     def raw_connect(self):
         raw_connection = psycopg2.connect(self._dsn)
@@ -314,9 +333,13 @@ class Postgres(Database):
         if self._version is None:
             cursor = raw_connection.cursor()
 
-            cursor.execute("SHOW server_version")
-            server_version = cursor.fetchone()[0]
-            self._version = tuple(map(int, server_version.split(".")))
+            try:
+                cursor.execute("SHOW server_version_num")
+            except psycopg2.ProgrammingError:
+                self._version = 0
+                raw_connection.rollback()
+            else:
+                self._version = int(cursor.fetchone()[0])
 
             # This will conditionally change the compilation of binary
             # variables (RawStrVariable) to preceed the placeholder with an
@@ -340,8 +363,7 @@ class Postgres(Database):
             raw_connection.rollback()
 
         raw_connection.set_client_encoding("UTF8")
-        raw_connection.set_isolation_level(
-            psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE)
+        raw_connection.set_isolation_level(self._isolation)
         return raw_connection
 
 
@@ -365,3 +387,18 @@ def make_dsn(uri):
     if uri.password is not None:
         dsn += " password=%s" % uri.password
     return dsn
+
+
+class PostgresTimeoutTracer(TimeoutTracer):
+
+    def set_statement_timeout(self, raw_cursor, remaining_time):
+        raw_cursor.execute("SET statement_timeout TO %d" %
+                           (remaining_time * 1000))
+
+    def connection_raw_execute_error(self, connection, raw_cursor,
+                                     statement, params, error):
+        # This should just check for
+        # psycopg2.extensions.QueryCanceledError in the future.
+        if (isinstance(error, DatabaseError) and
+            "statement timeout" in str(error)):
+            raise TimeoutError(statement, params)
