@@ -24,20 +24,22 @@
 This module contains the highest-level ORM interface in Storm.
 """
 
-
+from copy import copy
 from weakref import WeakValueDictionary
+from operator import itemgetter
 
 from storm.info import get_cls_info, get_obj_info, set_obj_info
 from storm.variables import Variable, LazyValue
 from storm.expr import (
     Expr, Select, Insert, Update, Delete, Column, Count, Max, Min,
     Avg, Sum, Eq, And, Asc, Desc, compile_python, compare_columns, SQLRaw,
-    Union, Except, Intersect, Alias)
+    Union, Except, Intersect, Alias, SetExpr)
 from storm.exceptions import (
     WrongStoreError, NotFlushedError, OrderLoopError, UnorderedError,
     NotOneError, FeatureError, CompileError, LostObjectError, ClassInfoError)
 from storm import Undef
 from storm.cache import Cache
+from storm.event import EventSystem
 
 
 __all__ = ["Store", "AutoReload", "EmptyResultSet"]
@@ -61,15 +63,23 @@ class Store(object):
 
     _result_set_factory = None
 
-    def __init__(self, database):
+    def __init__(self, database, cache=None):
         """
         @param database: The L{storm.database.Database} instance to use.
+        @param cache: The cache to use.  Defaults to a L{Cache} storing up to
+            100 objects at any given time.
         """
-        self._connection = database.connect()
+        self._event = EventSystem(self)
+        self._connection = database.connect(self._event)
         self._alive = WeakValueDictionary()
         self._dirty = {}
         self._order = {} # (info, info) = count
-        self._cache = Cache(100)
+        if cache is None:
+            self._cache = Cache(100)
+        else:
+            self._cache = cache
+        self._implicit_flush_block_count = 0
+        self._sequence = 0 # Advisory ordering.
 
     @staticmethod
     def of(obj):
@@ -89,7 +99,8 @@ class Store(object):
         This is just like L{storm.database.Database.execute}, except
         that a flush is performed first.
         """
-        self.flush()
+        if self._implicit_flush_block_count == 0:
+            self.flush()
         return self._connection.execute(statement, params, noresult)
 
     def close(self):
@@ -134,7 +145,8 @@ class Store(object):
             if no object is found.
         """
 
-        self.flush()
+        if self._implicit_flush_block_count == 0:
+            self.flush()
 
         if type(key) != tuple:
             key = (key,)
@@ -149,7 +161,8 @@ class Store(object):
                 variable = column.variable_factory(value=variable)
             primary_vars.append(variable)
 
-        obj_info = self._alive.get((cls_info.cls, tuple(primary_vars)))
+        primary_values = tuple(var.get(to_db=True) for var in primary_vars)
+        obj_info = self._alive.get((cls_info.cls, primary_values))
         if obj_info is not None:
             if obj_info.get("invalidated"):
                 try:
@@ -189,14 +202,11 @@ class Store(object):
         @return: A L{ResultSet} of instances C{cls_spec}. If C{cls_spec}
             was a tuple, then an iterator of tuples of such instances.
         """
-        self.flush()
-        if type(cls_spec) is tuple:
-            cls_spec_info = tuple(get_cls_info(cls) for cls in cls_spec)
-            where = get_where_for_args(args, kwargs)
-        else:
-            cls_spec_info = get_cls_info(cls_spec)
-            where = get_where_for_args(args, kwargs, cls_spec)
-        return self._result_set_factory(self, cls_spec_info, where)
+        if self._implicit_flush_block_count == 0:
+            self.flush()
+        find_spec = FindSpec(cls_spec)
+        where = get_where_for_args(args, kwargs, find_spec.default_cls)
+        return self._result_set_factory(self, find_spec, where)
 
     def using(self, *tables):
         """Specify tables to use explicitly.
@@ -231,6 +241,7 @@ class Store(object):
 
         The C{added} event will be fired on the object info's event system.
         """
+        self._event.emit("register-transaction")
         obj_info = get_obj_info(obj)
 
         store = obj_info.get("store")
@@ -259,6 +270,7 @@ class Store(object):
 
         The associated row will be deleted from the database.
         """
+        self._event.emit("register-transaction")
         obj_info = get_obj_info(obj)
 
         if obj_info.get("store") is not self:
@@ -273,10 +285,12 @@ class Store(object):
             del obj_info["pending"]
             self._set_clean(obj_info)
             self._disable_lazy_resolving(obj_info)
+            obj_info.event.emit("removed")
         else:
             obj_info["pending"] = PENDING_REMOVE
             self._set_dirty(obj_info)
             self._disable_lazy_resolving(obj_info)
+            obj_info.event.emit("removed")
 
     def reload(self, obj):
         """Reload the given object.
@@ -333,6 +347,28 @@ class Store(object):
             self._cache.remove(get_obj_info(obj))
         self._mark_autoreload(obj, True)
 
+    def reset(self):
+        """Reset this store, causing all future queries to return new objects.
+
+        Beware this method: it breaks the assumption that there will never be
+        two objects in memory which represent the same database object.
+
+        This is useful if you've got in-memory changes to an object that you
+        want to "throw out"; next time they're fetched the objects will be
+        recreated, so in-memory modifications will not be in effect for future
+        queries.
+        """
+        for obj_info in self._iter_alive():
+            if "store" in obj_info:
+                del obj_info["store"]
+        self._alive.clear()
+        self._dirty.clear()
+        self._cache.clear()
+        # The following line is untested, but then, I can't really find a way
+        # to test it without whitebox.
+        self._order.clear()
+
+
     def _mark_autoreload(self, obj=None, invalidate=False):
         if obj is None:
             obj_infos = self._iter_alive()
@@ -382,10 +418,7 @@ class Store(object):
             call to L{add_flush_order}.
         """
         pair = (get_obj_info(before), get_obj_info(after))
-        try:
-            self._order[pair] -= 1
-        except KeyError:
-            pass
+        self._order[pair] -= 1
 
     def flush(self):
         """Flush all dirty objects in cache to database.
@@ -406,8 +439,7 @@ class Store(object):
         normal flushing times are insufficient, such as when you want to
         make sure a database trigger gets run at a particular time.
         """
-        for obj_info in self._iter_alive():
-            obj_info.event.emit("flush")
+        self._event.emit("flush")
 
         # The _dirty list may change under us while we're running
         # the flush hooks, so we cannot just simply loop over it
@@ -430,20 +462,34 @@ class Store(object):
                 else:
                     before_set.add(before_info)
 
+        key_func = itemgetter("sequence")
+
+        # The external loop is important because items can get into the dirty
+        # state while we're flushing objects, ...
         while self._dirty:
-            for obj_info in self._dirty:
-                for before_info in predecessors.get(obj_info, ()):
-                    if before_info in self._dirty:
-                        break # A predecessor is still dirty.
+            # ... but we don't have to resort everytime an object is flushed,
+            # so we have an internal loop too.  If no objects become dirty
+            # during flush, this will clean self._dirty and the external loop
+            # will exit too.
+            sorted_dirty = sorted(self._dirty, key=key_func)
+            while sorted_dirty:
+                for i, obj_info in enumerate(sorted_dirty):
+                    for before_info in predecessors.get(obj_info, ()):
+                        if before_info in self._dirty:
+                            break # A predecessor is still dirty.
+                    else:
+                        break # Found an item without dirty predecessors.
                 else:
-                    break # Found an item without dirty predecessors.
-            else:
-                raise OrderLoopError("Can't flush due to ordering loop")
-            self._dirty.pop(obj_info, None)
-            self._flush_one(obj_info)
+                    raise OrderLoopError("Can't flush due to ordering loop")
+                del sorted_dirty[i]
+                self._dirty.pop(obj_info, None)
+                self._flush_one(obj_info)
 
         self._order.clear()
 
+        # That's not stricly necessary, but prevents getting into bigints.
+        self._sequence = 0
+
     def _flush_one(self, obj_info):
         cls_info = obj_info.cls_info
 
@@ -512,6 +558,15 @@ class Store(object):
 
         obj_info.event.emit("flushed")
 
+    def block_implicit_flushes(self):
+        """Block implicit flushes from operations like execute()."""
+        self._implicit_flush_block_count += 1
+
+    def unblock_implicit_flushes(self):
+        """Unblock implicit flushes from operations like execute()."""
+        assert self._implicit_flush_block_count > 0
+        self._implicit_flush_block_count -= 1
+
     def _get_changes_map(self, obj_info, adding=False):
         """Return a {column: variable} dictionary suitable for inserts/updates.
 
@@ -603,20 +658,6 @@ class Store(object):
             raise LostObjectError("Object is not in the database anymore")
         obj_info.pop("invalidated", None)
 
-    def _load_objects(self, cls_spec_info, result, values):
-        if type(cls_spec_info) is not tuple:
-            return self._load_object(cls_spec_info, result, values)
-        else:
-            objects = []
-            values_start = values_end = 0
-            for cls_info in cls_spec_info:
-                values_end += len(cls_info.columns)
-                obj = self._load_object(cls_info, result,
-                                        values[values_start:values_end])
-                objects.append(obj)
-                values_start = values_end
-            return tuple(objects)
-
     def _load_object(self, cls_info, result, values):
         # _set_values() need the cls_info columns for the class of the
         # actual object, not from a possible wrapper (e.g. an alias).
@@ -626,22 +667,24 @@ class Store(object):
         # Prepare cache key.
         primary_vars = []
         columns = cls_info.columns
-        is_null = True
+
+        for value in values:
+            if value is not None:
+                break
+        else:
+            # We've got a row full of NULLs, so consider that the object
+            # wasn't found.  This is useful for joins, where non-existent
+            # rows are represented like that.
+            return None
+
         for i in cls_info.primary_key_pos:
             value = values[i]
-            if value is not None:
-                is_null = False
             variable = columns[i].variable_factory(value=value, from_db=True)
             primary_vars.append(variable)
 
-        if is_null:
-            # We've got a row full of NULLs, so consider that the object
-            # wasn't found.  This is useful for joins, where unexistent
-            # rows are reprsented like that.
-            return None
-
         # Lookup cache.
-        obj_info = self._alive.get((cls, tuple(primary_vars)))
+        primary_values = tuple(var.get(to_db=True) for var in primary_vars)
+        obj_info = self._alive.get((cls, primary_values))
 
         if obj_info is not None:
             # Found object in cache, and it must be valid since the
@@ -683,6 +726,9 @@ class Store(object):
             obj = cls.__new__(cls)
             obj_info.set_obj(obj)
             set_obj_info(obj, obj_info)
+            # Re-enable change notification, as it may have been implicitely
+            # disabled when the previous object has been collected
+            self._enable_change_notification(obj_info)
             self._run_hook(obj_info, "__storm_loaded__")
         # Renew the cache.
         self._cache.add(obj_info)
@@ -725,7 +771,9 @@ class Store(object):
         return obj_info in self._dirty
 
     def _set_dirty(self, obj_info):
-        self._dirty[obj_info] = obj_info.get_obj()
+        if obj_info not in self._dirty:
+            self._dirty[obj_info] = obj_info.get_obj()
+            obj_info["sequence"] = self._sequence = self._sequence + 1
 
     def _set_clean(self, obj_info):
         self._dirty.pop(obj_info, None)
@@ -754,10 +802,14 @@ class Store(object):
         cls_info = obj_info.cls_info
         old_primary_vars = obj_info.get("primary_vars")
         if old_primary_vars is not None:
-            self._alive.pop((cls_info.cls, old_primary_vars), None)
+            old_primary_values = tuple(
+                var.get(to_db=True) for var in old_primary_vars)
+            self._alive.pop((cls_info.cls, old_primary_values), None)
         new_primary_vars = tuple(variable.copy()
                                  for variable in obj_info.primary_vars)
-        self._alive[cls_info.cls, new_primary_vars] = obj_info
+        new_primary_values = tuple(
+            var.get(to_db=True) for var in new_primary_vars)
+        self._alive[cls_info.cls, new_primary_values] = obj_info
         obj_info["primary_vars"] = new_primary_vars
         self._cache.add(obj_info)
 
@@ -771,18 +823,20 @@ class Store(object):
         primary_vars = obj_info.get("primary_vars")
         if primary_vars is not None:
             self._cache.remove(obj_info)
-            del self._alive[obj_info.cls_info.cls, primary_vars]
+            primary_values = tuple(var.get(to_db=True) for var in primary_vars)
+            del self._alive[obj_info.cls_info.cls, primary_values]
             del obj_info["primary_vars"]
 
     def _iter_alive(self):
         return self._alive.values()
 
-
     def _enable_change_notification(self, obj_info):
+        obj_info.event.emit("start-tracking-changes", self._event)
         obj_info.event.hook("changed", self._variable_changed)
 
     def _disable_change_notification(self, obj_info):
         obj_info.event.unhook("changed", self._variable_changed)
+        obj_info.event.emit("stop-tracking-changes", self._event)
 
     def _variable_changed(self, obj_info, variable,
                           old_value, new_value, fromdb):
@@ -813,11 +867,16 @@ class Store(object):
         the store, and then set all variables set to AutoReload to
         their database values.
         """
+        if lazy_value is not AutoReload and not isinstance(lazy_value, Expr):
+            # It's not something we handle.
+            return
+
         # XXX This will do it for now, but it should really flush
         #     just this single object and ones that it depends on.
         #     _flush_one() doesn't consider dependencies, so it may
         #     not be used directly.  Maybe allow flush(obj)?
-        self.flush()
+        if self._implicit_flush_block_count == 0:
+            self.flush()
 
         autoreload_columns = []
         for column in obj_info.cls_info.columns:
@@ -844,17 +903,19 @@ class ResultSet(object):
     Generally these should not be constructed directly, but instead
     retrieved from calls to L{Store.find}.
     """
-    def __init__(self, store, cls_spec_info,
+    def __init__(self, store, find_spec,
                  where=Undef, tables=Undef, select=Undef):
         self._store = store
-        self._cls_spec_info = cls_spec_info
+        self._find_spec = find_spec
         self._where = where
         self._tables = tables
         self._select = select
-        self._order_by = getattr(cls_spec_info, "default_order", Undef)
+        self._order_by = find_spec.default_order
         self._offset = Undef
         self._limit = Undef
         self._distinct = False
+        self._group_by = Undef
+        self._having = Undef
 
     def copy(self):
         """Return a copy of this ResultSet object, with the same configuration.
@@ -893,21 +954,14 @@ class ResultSet(object):
             if self._offset is not Undef: # XXX UNTESTED!
                 self._select.offset = self._offset
             return self._select
-        if type(self._cls_spec_info) is tuple:
-            columns = []
-            default_tables = []
-            for cls_info in self._cls_spec_info:
-                columns.append(cls_info.columns)
-                default_tables.append(cls_info.table)
-        else:
-            columns = self._cls_spec_info.columns
-            default_tables = self._cls_spec_info.table
+        columns, default_tables = self._find_spec.get_columns_and_tables()
         return Select(columns, self._where, self._tables, default_tables,
                       self._order_by, offset=self._offset, limit=self._limit,
-                      distinct=self._distinct)
+                      distinct=self._distinct, group_by=self._group_by,
+                      having=self._having)
 
     def _load_objects(self, result, values):
-        return self._store._load_objects(self._cls_spec_info, result, values)
+        return self._find_spec.load_objects(self._store, result, values)
 
     def __iter__(self):
         """Iterate the results of the query.
@@ -962,6 +1016,37 @@ class ResultSet(object):
 
         return self.copy().config(offset=offset, limit=limit)
 
+    def __contains__(self, item):
+        """Check if an item is contained within the result set."""
+        columns, values = self._find_spec.get_columns_and_values_for_item(item)
+
+        if self._select is Undef and self._group_by is Undef:
+            # No predefined select: adjust the where clause.
+            dummy, default_tables = self._find_spec.get_columns_and_tables()
+            where = [Eq(*pair) for pair in zip(columns, values)]
+            if self._where is not Undef:
+                where.append(self._where)
+            select = Select(1, And(*where), self._tables,
+                            default_tables)
+        else:
+            # Rewrite the predefined query and use it as a subquery.
+            aliased_columns = [Alias(column, "_key%d" % index)
+                               for (index, column) in enumerate(columns)]
+            subquery = replace_columns(self._get_select(), aliased_columns)
+            where = [Eq(*pair) for pair in zip(aliased_columns, values)]
+            select = Select(1, And(*where), Alias(subquery, "_tmp"))
+
+        result = self._store._connection.execute(select)
+        return result.get_one() is not None
+
+    def is_empty(self):
+        """Return true if this L{ResultSet} contains no results."""
+        subselect = self._get_select()
+        subselect.limit = 1
+        select = Select(1, tables=Alias(subselect, "_tmp"), limit=1)
+        result = self._store._connection.execute(select)
+        return (not result.get_one())
+
     def any(self):
         """Return a single item from the result set.
 
@@ -1055,23 +1140,55 @@ class ResultSet(object):
         This is done efficiently with a DELETE statement, so objects
         are not actually loaded into Python.
         """
+        if self._group_by is not Undef:
+            raise FeatureError("Removing isn't supported after a "
+                               " GROUP BY clause ")
         if self._offset is not Undef or self._limit is not Undef:
             raise FeatureError("Can't remove a sliced result set")
-        if type(self._cls_spec_info) is tuple:
-            raise FeatureError("Removing not yet supported with tuple finds")
+        if self._find_spec.default_cls_info is None:
+            raise FeatureError("Removing not yet supported for tuple or "
+                               "expression finds")
         if self._select is not Undef:
             raise FeatureError("Removing isn't supported with "
                                "set expressions (unions, etc)")
-        self._store._connection.execute(Delete(self._where,
-                                               self._cls_spec_info.table),
-                                        noresult=True)
+        self._store._connection.execute(
+            Delete(self._where, self._find_spec.default_cls_info.table),
+            noresult=True)
+
+    def group_by(self, *expr):
+        """Group this ResultSet by the given expressions.
+
+        @param expr: The expressions used in the GROUP BY statement.
+
+        @return: self (not a copy).
+        """
+        if self._select is not Undef:
+            raise FeatureError("Grouping isn't supported with "
+                               "set expressions (unions, etc)")
+
+        find_spec = FindSpec(expr)
+        columns, dummy = find_spec.get_columns_and_tables()
+
+        self._group_by = columns
+        return self
+
+    def having(self, *expr):
+        """Filter result previously grouped by.
+
+        @param expr: Instances of L{Expr}.
+
+        @return: self (not a copy).
+        """
+        if self._group_by is Undef:
+            raise FeatureError("having can only be called after group_by.")
+        self._having = And(*expr)
+        return self
 
     def _aggregate(self, expr, column=None):
-        if type(self._cls_spec_info) is tuple:
-            default_tables = [cls_info.table
-                              for cls_info in self._cls_spec_info]
-        else:
-            default_tables = self._cls_spec_info.table
+        if self._group_by is not Undef:
+            raise FeatureError("Single aggregates aren't supported after a "
+                               " GROUP BY clause ")
+        dummy, default_tables = self._find_spec.get_columns_and_tables()
         if self._select is Undef:
             select = Select(expr, self._where, self._tables, default_tables)
         else:
@@ -1087,7 +1204,16 @@ class ResultSet(object):
 
     def count(self, expr=Undef, distinct=False):
         """Get the number of objects represented by this ResultSet."""
-        return int(self._aggregate(Count(expr, distinct)))
+        if (self._distinct or self._limit is not Undef or
+            self._offset is not Undef):
+            subselect = self._get_select()
+            if expr is not Undef:
+                subselect.columns = expr
+            select = Select(Count(), tables=Alias(subselect, "_tmp"))
+            result = self._store._connection.execute(select)
+            return int(result.get_one()[0])
+        else:
+            return int(self._aggregate(Count(expr, distinct)))
 
     def max(self, expr):
         """Get the highest value from an expression."""
@@ -1147,9 +1273,13 @@ class ResultSet(object):
         For instance, C{result.set(Class.attr1 == 1, attr2=2)} will set
         C{attr1} to 1 and C{attr2} to 2, on all matching objects.
         """
+        if self._group_by is not Undef:
+            raise FeatureError("Setting isn't supported after a "
+                               " GROUP BY clause ")
 
-        if type(self._cls_spec_info) is tuple:
-            raise FeatureError("Setting isn't supported with tuple finds")
+        if self._find_spec.default_cls_info is None:
+            raise FeatureError("Setting isn't supported with tuple or "
+                               "expression finds")
         if self._select is not Undef:
             raise FeatureError("Setting isn't supported with "
                                "set expressions (unions, etc)")
@@ -1158,7 +1288,7 @@ class ResultSet(object):
             return
 
         changes = {}
-        cls = self._cls_spec_info.cls
+        cls = self._find_spec.default_cls_info.cls
 
         # For now only "Class.attr == var" is supported in args.
         for expr in args:
@@ -1181,7 +1311,8 @@ class ResultSet(object):
             else:
                 changes[column] = column.variable_factory(value=value)
 
-        expr = Update(changes, self._where, self._cls_spec_info.table)
+        expr = Update(changes, self._where,
+                      self._find_spec.default_cls_info.table)
         self._store.execute(expr, noresult=True)
 
         try:
@@ -1206,23 +1337,25 @@ class ResultSet(object):
 
     def cached(self):
         """Return matching objects from the cache for the current query."""
-        if type(self._cls_spec_info) is tuple:
-            raise FeatureError("Cache finds not supported with tuples")
+        if self._find_spec.default_cls_info is None:
+            raise FeatureError("Cache finds not supported with tuples "
+                               "or expressions")
         if self._tables is not Undef:
             raise FeatureError("Cache finds not supported with custom tables")
         if self._where is Undef:
             match = None
         else:
             match = compile_python.get_matcher(self._where)
-            name_to_column = dict((column.name, column)
-                                  for column in self._cls_spec_info.columns)
+            name_to_column = dict(
+                (column.name, column)
+                for column in self._find_spec.default_cls_info.columns)
             def get_column(name, name_to_column=name_to_column):
                 return obj_info.variables[name_to_column[name]].get()
         objects = []
-        cls = self._cls_spec_info.cls
+        cls = self._find_spec.default_cls_info.cls
         for obj_info in self._store._iter_alive():
             try:
-                if (obj_info.cls_info is self._cls_spec_info and
+                if (obj_info.cls_info is self._find_spec.default_cls_info and
                     (match is None or match(get_column))):
                     objects.append(self._store._get_object(obj_info))
             except LostObjectError:
@@ -1231,11 +1364,11 @@ class ResultSet(object):
         return objects
 
     def _set_expr(self, expr_cls, other, all=False):
-        if self._cls_spec_info != other._cls_spec_info:
+        if not self._find_spec.is_compatible(other._find_spec):
             raise FeatureError("Incompatible results for set operation")
 
         expr = expr_cls(self._get_select(), other._get_select(), all=all)
-        return ResultSet(self._store, self._cls_spec_info, select=expr)
+        return ResultSet(self._store, self._find_spec, select=expr)
 
     def union(self, other, all=False):
         """Get the L{Union} of this result set and another.
@@ -1302,6 +1435,12 @@ class EmptyResultSet(object):
     def __getitem__(self, index):
         return self.copy()
 
+    def __contains__(self, item):
+        return False
+
+    def is_empty(self):
+        return True
+
     def any(self):
         return None
 
@@ -1325,7 +1464,7 @@ class EmptyResultSet(object):
     def remove(self):
         pass
 
-    def count(self, column=Undef, distinct=False):
+    def count(self, expr=Undef, distinct=False):
         return 0
 
     def max(self, column):
@@ -1383,14 +1522,11 @@ class TableSet(object):
 
         @return: A L{ResultSet}.
         """
-        self._store.flush()
-        if type(cls_spec) is tuple:
-            cls_spec_info = tuple(get_cls_info(cls) for cls in cls_spec)
-            where = get_where_for_args(args, kwargs)
-        else:
-            cls_spec_info = get_cls_info(cls_spec)
-            where = get_where_for_args(args, kwargs, cls_spec)
-        return self._store._result_set_factory(self._store, cls_spec_info,
+        if self._store._implicit_flush_block_count == 0:
+            self._store.flush()
+        find_spec = FindSpec(cls_spec)
+        where = get_where_for_args(args, kwargs, find_spec.default_cls)
+        return self._store._result_set_factory(self._store, find_spec,
                                                where, self._tables)
 
 
@@ -1398,6 +1534,108 @@ Store._result_set_factory = ResultSet
 Store._table_set = TableSet
 
 
+class FindSpec(object):
+    """The set of tables or expressions in the result of L{Store.find}."""
+
+    def __init__(self, cls_spec):
+        self.is_tuple = type(cls_spec) == tuple
+        if not self.is_tuple:
+            cls_spec = (cls_spec,)
+
+        info = []
+        for item in cls_spec:
+            if isinstance(item, Expr):
+                info.append((True, item))
+            else:
+                info.append((False, get_cls_info(item)))
+        self._cls_spec_info = tuple(info)
+
+        # Do we have a single non-expression item here?
+        if not self.is_tuple and not info[0][0]:
+            self.default_cls = cls_spec[0]
+            self.default_cls_info = info[0][1]
+            self.default_order = self.default_cls_info.default_order
+        else:
+            self.default_cls = None
+            self.default_cls_info = None
+            self.default_order = Undef
+
+    def get_columns_and_tables(self):
+        columns = []
+        default_tables = []
+        for is_expr, info in self._cls_spec_info:
+            if is_expr:
+                columns.append(info)
+                if isinstance(info, Column):
+                    default_tables.append(info.table)
+            else:
+                columns.extend(info.columns)
+                default_tables.append(info.table)
+        return columns, default_tables
+
+    def is_compatible(self, find_spec):
+        """Return True if this FindSpec is compatible with a second one."""
+        if self.is_tuple != find_spec.is_tuple:
+            return False
+        if len(self._cls_spec_info) != len(find_spec._cls_spec_info):
+            return False
+        for (is_expr1, info1), (is_expr2, info2) in zip(
+            self._cls_spec_info, find_spec._cls_spec_info):
+            if is_expr1 != is_expr2:
+                return False
+            if info1 is not info2:
+                return False
+        return True
+
+    def load_objects(self, store, result, values):
+        objects = []
+        values_start = values_end = 0
+        for is_expr, info in self._cls_spec_info:
+            if is_expr:
+                values_end += 1
+                variable = getattr(info, "variable_factory", Variable)(
+                    value=values[values_start], from_db=True)
+                objects.append(variable.get())
+            else:
+                values_end += len(info.columns)
+                obj = store._load_object(info, result,
+                                         values[values_start:values_end])
+                objects.append(obj)
+            values_start = values_end
+        if self.is_tuple:
+            return tuple(objects)
+        else:
+            return objects[0]
+
+    def get_columns_and_values_for_item(self, item):
+        """Generate a comparison expression with the given item."""
+        if isinstance(item, tuple):
+            if not self.is_tuple:
+                raise TypeError("Find spec does not expect tuples.")
+        else:
+            if self.is_tuple:
+                raise TypeError("Find spec expects tuples.")
+            item = (item,)
+
+        columns = []
+        values = []
+        for (is_expr, info), value in zip(self._cls_spec_info, item):
+            if is_expr:
+                if not isinstance(value, (Expr, Variable)) and (
+                    value is not None):
+                    value = getattr(info, "variable_factory", Variable)(
+                        value=value)
+                columns.append(info)
+                values.append(value)
+            else:
+                obj_info = get_obj_info(value)
+                if obj_info.cls_info != info:
+                    raise TypeError("%r does not match %r" % (value, info))
+                columns.extend(info.primary_key)
+                values.extend(obj_info.primary_vars)
+        return columns, values
+
+
 def get_where_for_args(args, kwargs, cls=None):
     equals = list(args)
     if kwargs:
@@ -1411,6 +1649,34 @@ def get_where_for_args(args, kwargs, cls=None):
     return Undef
 
 
+def replace_columns(expr, columns):
+    if isinstance(expr, Select):
+        select = copy(expr)
+        select.columns = columns
+        # Remove the ordering if it won't affect the result of the query.
+        if select.limit is Undef and select.offset is Undef:
+            select.order_by = Undef
+        return select
+    elif isinstance(expr, SetExpr):
+        # The ORDER BY clause might refer to columns we have replaced.
+        # Luckily we can ignore it if there is no limit/offset.
+        if expr.order_by is not Undef and (
+            expr.limit is not Undef or expr.offset is not Undef):
+            raise FeatureError(
+                "__contains__() does not yet support set "
+                "expressions that combine ORDER BY with "
+                "LIMIT/OFFSET")
+        subexprs = [replace_columns(subexpr, columns)
+                    for subexpr in expr.exprs]
+        return expr.__class__(
+            all=expr.all, limit=expr.limit, offset=expr.offset,
+            *subexprs)
+    else:
+        raise FeatureError(
+            "__contains__() does not yet support %r expressions"
+            % (expr.__class__,))
+
+
 class AutoReload(LazyValue):
     """A marker for reloading a single value.
 
