@@ -26,10 +26,12 @@ from storm.exceptions import ClosedError, DatabaseError, DisconnectionError
 from storm.variables import Variable
 import storm.database
 from storm.database import *
+from storm.tracer import install_tracer, remove_all_tracers, DebugTracer
 from storm.uri import URI
 from storm.expr import *
 
 from tests.helper import TestHelper
+from tests.mocker import ARGS
 
 
 marker = object()
@@ -96,6 +98,27 @@ class FakeConnection(object):
         return _function(*args, **kwargs)
 
 
+class FakeTracer(object):
+
+    def __init__(self, stream=None):
+        self.seen = []
+
+    def connection_raw_execute(self, connection, raw_cursor,
+                               statement, params):
+        self.seen.append(("EXECUTE", connection, type(raw_cursor),
+                          statement, params))
+
+    def connection_raw_execute_success(self, connection, raw_cursor,
+                                       statement, params):
+        self.seen.append(("SUCCESS", connection, type(raw_cursor),
+                          statement, params))
+
+    def connection_raw_execute_error(self, connection, raw_cursor,
+                                     statement, params, error):
+        self.seen.append(("ERROR", connection, type(raw_cursor),
+                          statement, params, error))
+
+
 class DatabaseTest(TestHelper):
 
     def setUp(self):
@@ -116,6 +139,10 @@ class ConnectionTest(TestHelper):
         self.database.raw_connect = lambda: self.raw_connection
         self.connection = Connection(self.database)
 
+    def tearDown(self):
+        TestHelper.tearDown(self)
+        remove_all_tracers()
+
     def test_execute(self):
         result = self.connection.execute("something")
         self.assertTrue(isinstance(result, Result))
@@ -166,6 +193,40 @@ class ConnectionTest(TestHelper):
         self.connection.close()
         self.assertRaises(ClosedError, self.connection.execute, "SELECT 1")
 
+    def test_raw_execute_tracing(self):
+        self.assertMethodsMatch(FakeTracer, DebugTracer)
+        tracer = FakeTracer()
+        install_tracer(tracer)
+        self.connection.execute("something")
+        self.assertEquals(tracer.seen, [("EXECUTE", self.connection, RawCursor,
+                                         "something", ()),
+                                        ("SUCCESS", self.connection, RawCursor,
+                                         "something", ())])
+
+        del tracer.seen[:]
+        self.connection.execute("something", (1, 2))
+        self.assertEquals(tracer.seen, [("EXECUTE", self.connection, RawCursor,
+                                         "something", (1, 2)),
+                                        ("SUCCESS", self.connection, RawCursor,
+                                         "something", (1, 2))])
+
+    def test_raw_execute_error_tracing(self):
+        cursor_mock = self.mocker.patch(RawCursor)
+        cursor_mock.execute(ARGS)
+        error = ZeroDivisionError()
+        self.mocker.throw(error)
+        self.mocker.replay()
+
+        self.assertMethodsMatch(FakeTracer, DebugTracer)
+        tracer = FakeTracer()
+        install_tracer(tracer)
+        self.assertRaises(ZeroDivisionError,
+                          self.connection.execute, "something")
+        self.assertEquals(tracer.seen, [("EXECUTE", self.connection, RawCursor,
+                                         "something", ()),
+                                        ("ERROR", self.connection, RawCursor,
+                                         "something", (), error)])
+
     def test_commit(self):
         self.connection.commit()
         self.assertEquals(self.executed, ["COMMIT"])
