@@ -31,6 +31,7 @@ from storm.variables import (Variable, PickleVariable, RawStrVariable,
                              DecimalVariable, DateTimeVariable, DateVariable,
                              TimeVariable, TimeDeltaVariable)
 from storm.database import *
+from storm.event import EventSystem
 from storm.exceptions import (
     DatabaseModuleError, DisconnectionError, OperationalError)
 
@@ -38,6 +39,12 @@ from tests.databases.proxy import ProxyTCPServer
 from tests.helper import MakePath
 
 
+class Marker(object):
+    pass
+
+marker = Marker()
+
+
 class DatabaseTest(object):
 
     supports_microseconds = True
@@ -336,6 +343,22 @@ class DatabaseTest(object):
         finally:
             connection1.rollback()
 
+    def test_wb_connect_sets_event_system(self):
+        connection = self.database.connect(marker)
+        self.assertEqual(connection._event, marker)
+
+    def test_execute_sends_event(self):
+        event = EventSystem(marker)
+        calls = []
+        def register_transaction(owner):
+            calls.append(owner)
+        event.hook("register-transaction", register_transaction)
+
+        connection = self.database.connect(event)
+        connection.execute("SELECT 1")
+        self.assertEqual(len(calls), 1)
+        self.assertEqual(calls[0], marker)
+
     def from_database(self, row):
         return [int(item)+1 for item in row]
 
@@ -525,3 +548,11 @@ class DatabaseDisconnectionTest(object):
         self.connection.rollback()
         self.assertRaises(DisconnectionError,
                           self.connection.execute, "SELECT 1")
+
+    def test_close_connection_after_disconnect(self):
+        result = self.connection.execute("SELECT 1")
+        self.assertTrue(result.get_one())
+        self.proxy.stop()
+        self.assertRaises(DisconnectionError,
+                          self.connection.execute, "SELECT 1")
+        self.connection.close()
