@@ -19,6 +19,7 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 from datetime import datetime, date, time, timedelta
+from decimal import Decimal as decimal
 import gc
 
 from storm.exceptions import NoneError, PropertyPathError
@@ -26,8 +27,9 @@ from storm.properties import PropertyPublisherMeta
 from storm.properties import *
 from storm.variables import *
 from storm.info import get_obj_info
-from storm.expr import Undef, Column, Select, compile, SQLRaw
+from storm.expr import Column, Select, compile, State, SQLRaw
 
+from tests.info import Wrapper
 from tests.helper import TestHelper
 
 
@@ -43,12 +45,12 @@ class PropertyTest(TestHelper):
     def setUp(self):
         TestHelper.setUp(self)
         class Class(object):
-            __storm_table__ = "table"
+            __storm_table__ = "mytable"
             prop1 = Custom("column1", primary=True)
             prop2 = Custom()
             prop3 = Custom("column3", default=50, allow_none=False)
         class SubClass(Class):
-            __storm_table__ = "subtable"
+            __storm_table__ = "mysubtable"
         self.Class = Class
         self.SubClass = SubClass
 
@@ -102,6 +104,24 @@ class PropertyTest(TestHelper):
         self.assertTrue(isinstance(variable, CustomVariable))
         self.assertTrue(variable.is_defined())
 
+    def test_variable_factory_validator_attribute(self):
+        # Should work even if we make things harder by reusing properties.
+        prop = Custom()
+        class Class1(object):
+            __storm_table__ = "table1"
+            prop1 = prop
+        class Class2(object):
+            __storm_table__ = "table2"
+            prop2 = prop
+        args = []
+        def validator(obj, attr, value):
+            args.append((obj, attr, value))
+        variable1 = Class1.prop1.variable_factory(validator=validator)
+        variable2 = Class2.prop2.variable_factory(validator=validator)
+        variable1.set(1)
+        variable2.set(2)
+        self.assertEquals(args, [(None, "prop1", 1), (None, "prop2", 2)])
+
     def test_default(self):
         obj = self.SubClass()
         self.assertEquals(obj.prop1, None)
@@ -125,6 +145,22 @@ class PropertyTest(TestHelper):
         self.assertEquals(obj.prop2, None)
         self.assertRaises(NoneError, setattr, obj, "prop3", None)
 
+    def test_set_with_validator(self):
+        args = []
+        def validator(obj, attr, value):
+            args[:] = obj, attr, value
+            return 42
+
+        class Class(object):
+            __storm_table__ = "mytable"
+            prop = Custom("column", primary=True, validator=validator)
+
+        obj = Class()
+        obj.prop = 21
+
+        self.assertEquals(args, [obj, "prop", 21])
+        self.assertEquals(obj.prop, 42)
+
     def test_set_get_subclass(self):
         obj = self.SubClass()
         obj.prop1 = 10
@@ -213,14 +249,18 @@ class PropertyTest(TestHelper):
         expr = Select(SQLRaw("*"), (prop1 == "value1") &
                                    (prop2 == "value2") &
                                    (prop3 == "value3"))
-        statement, parameters = compile(expr)
-        self.assertEquals(statement, "SELECT * FROM table WHERE "
-                                     "table.column1 = ? AND "
-                                     "table.prop2 = ? AND "
-                                     "table.column3 = ?")
-        self.assertEquals(parameters, [CustomVariable("value1"),
-                                       CustomVariable("value2"),
-                                       CustomVariable("value3")])
+        state = State()
+        statement = compile(expr, state)
+        self.assertEquals(statement, "SELECT * FROM mytable WHERE "
+                                     "mytable.column1 = ? AND "
+                                     "mytable.prop2 = ? AND "
+                                     "mytable.column3 = ?")
+
+        self.assertVariablesEqual(
+            state.parameters,
+            [CustomVariable("value1"),
+             CustomVariable("value2"),
+             CustomVariable("value3")])
 
     def test_comparable_expr_subclass(self):
         prop1 = self.SubClass.prop1
@@ -229,14 +269,45 @@ class PropertyTest(TestHelper):
         expr = Select(SQLRaw("*"), (prop1 == "value1") &
                                    (prop2 == "value2") &
                                    (prop3 == "value3"))
-        statement, parameters = compile(expr)
-        self.assertEquals(statement, "SELECT * FROM subtable WHERE "
-                                     "subtable.column1 = ? AND "
-                                     "subtable.prop2 = ? AND "
-                                     "subtable.column3 = ?")
-        self.assertEquals(parameters, [CustomVariable("value1"),
-                                       CustomVariable("value2"),
-                                       CustomVariable("value3")])
+        state = State()
+        statement = compile(expr, state)
+        self.assertEquals(statement, "SELECT * FROM mysubtable WHERE "
+                                     "mysubtable.column1 = ? AND "
+                                     "mysubtable.prop2 = ? AND "
+                                     "mysubtable.column3 = ?")
+        self.assertVariablesEqual(
+            state.parameters,
+            [CustomVariable("value1"),
+             CustomVariable("value2"),
+             CustomVariable("value3")])
+
+    def test_set_get_delete_with_wrapper(self):
+        obj = self.Class()
+        get_obj_info(obj) # Ensure the obj_info exists for obj.
+        self.Class.prop1.__set__(Wrapper(obj), 10)
+        self.assertEquals(self.Class.prop1.__get__(Wrapper(obj)), 10)
+        self.Class.prop1.__delete__(Wrapper(obj))
+        self.assertEquals(self.Class.prop1.__get__(Wrapper(obj)), None)
+
+    def test_reuse_of_instance(self):
+        """Properties are dynamically bound to the class where they're used.
+
+        It basically means that the property may be instantiated
+        independently from the class itself, and reused in any number of
+        classes.  It's not something we should announce as granted, but
+        right now it works, and we should try not to break it.
+        """
+        prop = Custom()
+        class Class1(object):
+            __storm_table__ = "table1"
+            prop1 = prop
+        class Class2(object):
+            __storm_table__ = "table2"
+            prop2 = prop
+        self.assertEquals(Class1.prop1.name, "prop1")
+        self.assertEquals(Class1.prop1.table, Class1)
+        self.assertEquals(Class2.prop2.name, "prop2")
+        self.assertEquals(Class2.prop2.table, Class2)
 
 
 class PropertyKindsTest(TestHelper):
@@ -245,7 +316,7 @@ class PropertyKindsTest(TestHelper):
         prop2_kwargs = kwargs.pop("prop2_kwargs", {})
         kwargs["primary"] = True
         class Class(object):
-            __storm_table__ = "table"
+            __storm_table__ = "mytable"
             prop1 = property("column1", *args, **kwargs)
             prop2 = property(**prop2_kwargs)
         class SubClass(Class):
@@ -323,8 +394,8 @@ class PropertyKindsTest(TestHelper):
         self.obj.prop1 = 1
         self.assertTrue(isinstance(self.obj.prop1, float))
 
-    def test_str(self):
-        self.setup(Chars, default="def", allow_none=False)
+    def test_decimal(self):
+        self.setup(Decimal, default=decimal("50.5"), allow_none=False)
 
         self.assertTrue(isinstance(self.column1, Column))
         self.assertTrue(isinstance(self.column2, Column))
@@ -332,8 +403,28 @@ class PropertyKindsTest(TestHelper):
         self.assertEquals(self.column1.table, self.SubClass)
         self.assertEquals(self.column2.name, "prop2")
         self.assertEquals(self.column2.table, self.SubClass)
-        self.assertTrue(isinstance(self.variable1, CharsVariable))
-        self.assertTrue(isinstance(self.variable2, CharsVariable))
+        self.assertTrue(isinstance(self.variable1, DecimalVariable))
+        self.assertTrue(isinstance(self.variable2, DecimalVariable))
+
+        self.assertEquals(self.obj.prop1, decimal("50.5"))
+        self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
+        self.obj.prop2 = None
+        self.assertEquals(self.obj.prop2, None)
+
+        self.obj.prop1 = 1
+        self.assertTrue(isinstance(self.obj.prop1, decimal))
+
+    def test_str(self):
+        self.setup(RawStr, default="def", allow_none=False)
+
+        self.assertTrue(isinstance(self.column1, Column))
+        self.assertTrue(isinstance(self.column2, Column))
+        self.assertEquals(self.column1.name, "column1")
+        self.assertEquals(self.column1.table, self.SubClass)
+        self.assertEquals(self.column2.name, "prop2")
+        self.assertEquals(self.column2.table, self.SubClass)
+        self.assertTrue(isinstance(self.variable1, RawStrVariable))
+        self.assertTrue(isinstance(self.variable2, RawStrVariable))
 
         self.assertEquals(self.obj.prop1, "def")
         self.assertRaises(NoneError, setattr, self.obj, "prop1", None)
@@ -547,6 +638,7 @@ class PropertyKindsTest(TestHelper):
         self.obj.prop2 = []
 
         self.obj_info.checkpoint()
+        self.obj_info.event.emit("start-tracking-changes", self.obj_info.event)
         self.obj_info.event.hook("changed", changed)
 
         self.assertEquals(self.obj.prop1, [])
@@ -596,6 +688,7 @@ class PropertyKindsTest(TestHelper):
             changes.append((variable, old_value, new_value, fromdb))
 
         self.obj_info.checkpoint()
+        self.obj_info.event.emit("start-tracking-changes", self.obj_info.event)
         self.obj_info.event.hook("changed", changed)
 
         self.assertEquals(self.obj.prop1, [])
@@ -618,12 +711,18 @@ class PropertyKindsTest(TestHelper):
     def test_variable_factory_arguments(self):
         class Class(object):
             __storm_table__ = "test"
+            id = Int(primary=True)
+
+        validator_args = []
+        def validator(obj, attr, value):
+            validator_args[:] = obj, attr, value
+            return value
 
         for func, cls, value in [
                                (Bool, BoolVariable, True),
                                (Int, IntVariable, 1),
                                (Float, FloatVariable, 1.1),
-                               (Chars, CharsVariable, "str"),
+                               (RawStr, RawStrVariable, "str"),
                                (Unicode, UnicodeVariable, u"unicode"),
                                (DateTime, DateTimeVariable, datetime.now()),
                                (Date, DateVariable, date.today()),
@@ -632,8 +731,8 @@ class PropertyKindsTest(TestHelper):
                                      ]:
 
             # Test no default and allow_none=True.
-            prop = func(name="name")
-            column = prop.__get__(None, Class)
+            Class.prop = func(name="name")
+            column = Class.prop.__get__(None, Class)
             self.assertEquals(column.name, "name")
             self.assertEquals(column.table, Class)
 
@@ -644,8 +743,8 @@ class PropertyKindsTest(TestHelper):
             self.assertEquals(variable.get(), None)
 
             # Test default and allow_none=False.
-            prop = func(name="name", default=value, allow_none=False)
-            column = prop.__get__(None, Class)
+            Class.prop = func(name="name", default=value, allow_none=False)
+            column = Class.prop.__get__(None, Class)
             self.assertEquals(column.name, "name")
             self.assertEquals(column.table, Class)
 
@@ -655,8 +754,8 @@ class PropertyKindsTest(TestHelper):
             self.assertEquals(variable.get(), value)
 
             # Test default_factory.
-            prop = func(name="name", default_factory=lambda:value)
-            column = prop.__get__(None, Class)
+            Class.prop = func(name="name", default_factory=lambda:value)
+            column = Class.prop.__get__(None, Class)
             self.assertEquals(column.name, "name")
             self.assertEquals(column.table, Class)
 
@@ -664,6 +763,21 @@ class PropertyKindsTest(TestHelper):
             self.assertTrue(isinstance(variable, cls))
             self.assertEquals(variable.get(), value)
 
+            # Test validator.
+            Class.prop = func(name="name", validator=validator, default=value)
+            column = Class.prop.__get__(None, Class)
+            self.assertEquals(column.name, "name")
+            self.assertEquals(column.table, Class)
+
+            del validator_args[:]
+            variable = column.variable_factory()
+            self.assertTrue(isinstance(variable, cls))
+            # Validator is not called on instantiation.
+            self.assertEquals(validator_args, [])
+            # But is when setting the variable.
+            variable.set(value)
+            self.assertEquals(validator_args, [None, "prop", value])
+
 
 class PropertyRegistryTest(TestHelper):
 
@@ -671,12 +785,12 @@ class PropertyRegistryTest(TestHelper):
         TestHelper.setUp(self)
 
         class Class(object):
-            __storm_table__ = "table"
+            __storm_table__ = "mytable"
             prop1 = Property("column1", primary=True)
             prop2 = Property()
 
         class SubClass(Class):
-            __storm_table__ = "subtable"
+            __storm_table__ = "mysubtable"
 
         self.Class = Class
         self.SubClass = SubClass
@@ -784,12 +898,12 @@ class PropertyPublisherMetaTest(TestHelper):
             __metaclass__ = PropertyPublisherMeta
 
         class Class(Base):
-            __storm_table__ = "table"
+            __storm_table__ = "mytable"
             prop1 = Property("column1", primary=True)
             prop2 = Property()
 
         class SubClass(Class):
-            __storm_table__ = "subtable"
+            __storm_table__ = "mysubtable"
 
         self.Class = Class
         self.SubClass = SubClass
