@@ -61,34 +61,42 @@ class PostgresStoreTest(TestHelper, StoreTest):
         self.database = create_database(os.environ["STORM_POSTGRES_URI"])
 
     def create_tables(self):
-        connection = self.database.connect()
+        connection = self.connection
         connection.execute("CREATE TABLE foo "
                            "(id SERIAL PRIMARY KEY,"
                            " title VARCHAR DEFAULT 'Default Title')")
+        # Prevent dynamically created Foos from having conflicting ids.
+        connection.execute("SELECT setval('foo_id_seq', 1000)")
         connection.execute("CREATE TABLE bar "
                            "(id SERIAL PRIMARY KEY,"
                            " foo_id INTEGER, title VARCHAR)")
         connection.execute("CREATE TABLE bin "
-                           "(id SERIAL PRIMARY KEY, bin BYTEA)")
+                           "(id SERIAL PRIMARY KEY, bin BYTEA, foo_id INTEGER)")
         connection.execute("CREATE TABLE link "
-                           "(foo_id INTEGER, bar_id INTEGER)")
+                           "(foo_id INTEGER, bar_id INTEGER,"
+                           " PRIMARY KEY (foo_id, bar_id))")
         connection.execute("CREATE TABLE money "
                            "(id SERIAL PRIMARY KEY, value NUMERIC(6,4))")
+        connection.execute("CREATE TABLE selfref "
+                           "(id SERIAL PRIMARY KEY, title VARCHAR,"
+                           " selfref_id INTEGER REFERENCES selfref(id))")
         connection.execute("CREATE TABLE lst1 "
                            "(id SERIAL PRIMARY KEY, ints INTEGER[])")
         connection.execute("CREATE TABLE lst2 "
                            "(id SERIAL PRIMARY KEY, ints INTEGER[][])")
+        connection.execute("CREATE TABLE foovalue "
+                           "(id SERIAL PRIMARY KEY, foo_id INTEGER,"
+                           " value1 INTEGER, value2 INTEGER)")
         connection.commit()
 
     def drop_tables(self):
         StoreTest.drop_tables(self)
         for table in ["lst1", "lst2"]:
-            connection = self.database.connect()
             try:
-                connection.execute("DROP TABLE %s" % table)
-                connection.commit()
+                self.connection.execute("DROP TABLE %s" % table)
+                self.connection.commit()
             except:
-                connection.rollback()
+                self.connection.rollback()
 
     def test_list_variable(self):
 
@@ -173,8 +181,7 @@ class PostgresEmptyResultSetTest(TestHelper, EmptyResultSetTest):
         self.database = create_database(os.environ["STORM_POSTGRES_URI"])
 
     def create_tables(self):
-        connection = self.database.connect()
-        connection.execute("CREATE TABLE foo "
-                           "(id SERIAL PRIMARY KEY,"
-                           " title VARCHAR DEFAULT 'Default Title')")
-        connection.commit()
+        self.connection.execute("CREATE TABLE foo "
+                                "(id SERIAL PRIMARY KEY,"
+                                " title VARCHAR DEFAULT 'Default Title')")
+        self.connection.commit()
