@@ -102,10 +102,6 @@ class VariableTest(TestHelper):
         self.assertEquals(variable.get(to_db=True), ("g", ("s", marker)))
         self.assertEquals(variable.gets, [(("s", marker), True)])
 
-    def test_eq(self):
-        self.assertEquals(CustomVariable(marker), CustomVariable(marker))
-        self.assertNotEquals(CustomVariable(marker), CustomVariable(object()))
-
     def test_is_defined(self):
         variable = CustomVariable()
         self.assertFalse(variable.is_defined())
@@ -141,6 +137,56 @@ class VariableTest(TestHelper):
             pass
         self.assertTrue("table_name.column_name" in str(e))
 
+    def test_set_with_validator(self):
+        args = []
+        def validator(obj, attr, value):
+            args.append((obj, attr, value))
+            return value
+        variable = CustomVariable(validator=validator)
+        variable.set(3)
+        self.assertEquals(args, [(None, None, 3)])
+
+    def test_set_with_validator_and_validator_arguments(self):
+        args = []
+        def validator(obj, attr, value):
+            args.append((obj, attr, value))
+            return value
+        variable = CustomVariable(validator=validator,
+                                  validator_object_factory=lambda: 1,
+                                  validator_attribute=2)
+        variable.set(3)
+        self.assertEquals(args, [(1, 2, 3)])
+
+    def test_set_with_validator_raising_error(self):
+        args = []
+        def validator(obj, attr, value):
+            args.append((obj, attr, value))
+            raise ZeroDivisionError()
+        variable = CustomVariable(validator=validator)
+        self.assertRaises(ZeroDivisionError, variable.set, marker)
+        self.assertEquals(args, [(None, None, marker)])
+        self.assertEquals(variable.get(), None)
+
+    def test_set_with_validator_changing_value(self):
+        args = []
+        def validator(obj, attr, value):
+            args.append((obj, attr, value))
+            return 42
+        variable = CustomVariable(validator=validator)
+        variable.set(marker)
+        self.assertEquals(args, [(None, None, marker)])
+        self.assertEquals(variable.get(), ('g', ('s', 42)))
+
+    def test_set_from_db_wont_call_validator(self):
+        args = []
+        def validator(obj, attr, value):
+            args.append((obj, attr, value))
+            return 42
+        variable = CustomVariable(validator=validator)
+        variable.set(marker, from_db=True)
+        self.assertEquals(args, [])
+        self.assertEquals(variable.get(), ('g', ('s', marker)))
+
     def test_event_changed(self):
         event = EventSystem(marker)
 
@@ -148,7 +194,7 @@ class VariableTest(TestHelper):
         def changed(owner, variable, old_value, new_value, fromdb):
             changed_values.append((owner, variable,
                                    old_value, new_value, fromdb))
-        
+
         event.hook("changed", changed)
 
         variable = CustomVariable(event=event)
@@ -210,14 +256,9 @@ class VariableTest(TestHelper):
         variable = CustomVariable()
         variable.set(marker)
         variable_copy = variable.copy()
+        variable_copy.gets = []
         self.assertTrue(variable is not variable_copy)
-        self.assertTrue(variable == variable_copy)
-
-    def test_hash(self):
-        # They must hash the same to be used as cache keys.
-        obj1 = CustomVariable(marker)
-        obj2 = CustomVariable(marker)
-        self.assertEquals(hash(obj1), hash(obj2))
+        self.assertVariablesEqual([variable], [variable_copy])
 
     def test_lazy_value_setting(self):
         variable = CustomVariable()
@@ -260,7 +301,7 @@ class VariableTest(TestHelper):
         def changed(owner, variable, old_value, new_value, fromdb):
             changed_values.append((owner, variable,
                                    old_value, new_value, fromdb))
-        
+
         event.hook("changed", changed)
 
         variable = CustomVariable(event=event)
@@ -601,7 +642,7 @@ class TimeDeltaVariableTest(TestHelper):
         self.assertEquals(variable.get(), delta)
 
         self.assertRaises(TypeError, variable.set, marker)
-    
+
     def test_get_set_from_database(self):
         delta_str = "42 days 12:34:56.78"
         delta_uni = unicode(delta_str)
@@ -739,7 +780,7 @@ class PickleVariableTest(TestHelper):
         self.assertEquals(variable.get(to_db=True), d_dump)
 
         self.assertEquals(variable.get_state(), (Undef, d_dump))
-        
+
         variable.set(marker)
         variable.set_state((Undef, d_dump))
         self.assertEquals(variable.get(), d)
@@ -756,6 +797,7 @@ class PickleVariableTest(TestHelper):
         def changed(owner, variable, old_value, new_value, fromdb):
             changes.append((variable, old_value, new_value, fromdb))
 
+        event.emit("start-tracking-changes", event)
         event.hook("changed", changed)
 
         variable.checkpoint()
@@ -793,13 +835,13 @@ class ListVariableTest(TestHelper):
 
         variable.set(l)
         self.assertEquals(variable.get(), l)
-        self.assertEquals(variable.get(to_db=True),
-                          [IntVariable(1), IntVariable(2)])
+        self.assertVariablesEqual(variable.get(to_db=True),
+                                  [IntVariable(1), IntVariable(2)])
 
         variable.set([1.1, 2.2], from_db=True)
         self.assertEquals(variable.get(), l)
-        self.assertEquals(variable.get(to_db=True),
-                          [IntVariable(1), IntVariable(2)])
+        self.assertVariablesEqual(variable.get(to_db=True),
+                                  [IntVariable(1), IntVariable(2)])
 
         self.assertEquals(variable.get_state(), (Undef, l_dump))
 
@@ -820,6 +862,7 @@ class ListVariableTest(TestHelper):
         def changed(owner, variable, old_value, new_value, fromdb):
             changes.append((variable, old_value, new_value, fromdb))
 
+        event.emit("start-tracking-changes", event)
         event.hook("changed", changed)
 
         variable.checkpoint()
